@@ -1,15 +1,16 @@
 #! /usr/bin/env/python
 # -*- coding: utf-8 -*-
+# cython: language_level=3
 
-## Copyright 1999-2012 by LivingLogic AG, Bayreuth/Germany
-## Copyright 1999-2012 by Walter Dörwald
+## Copyright 1999-2014 by LivingLogic AG, Bayreuth/Germany
+## Copyright 1999-2014 by Walter Dörwald
 ##
 ## All Rights Reserved
 ##
 ## See ll/xist/__init__.py for the license
 
 
-import sys, unittest, io
+import sys, io
 
 from xml.parsers import expat
 
@@ -26,39 +27,39 @@ import xist_common as common
 present.reprtab = "  "
 
 
-def check_lenunicode(node, _len, content):
+def check_lenstr(node, _len, content):
 	assert len(node) == _len
 	assert str(node) == content
 
 
 def test_fraglen():
-	check_lenunicode(xsc.Frag(), 0, "")
-	check_lenunicode(xsc.Frag(1), 1, "1")
-	check_lenunicode(xsc.Frag(1, 2, 3), 3, "123")
-	check_lenunicode(xsc.Frag(None), 0, "")
-	check_lenunicode(xsc.Frag(None, None, None), 0, "")
-	check_lenunicode(xsc.Frag(1, None, 2, None, 3, None, 4), 4, "1234")
-	check_lenunicode(xsc.Frag(1, (2, 3)), 3, "123")
-	check_lenunicode(xsc.Frag(1, (None, None)), 1, "1")
+	check_lenstr(xsc.Frag(), 0, "")
+	check_lenstr(xsc.Frag(1), 1, "1")
+	check_lenstr(xsc.Frag(1, 2, 3), 3, "123")
+	check_lenstr(xsc.Frag(None), 0, "")
+	check_lenstr(xsc.Frag(None, None, None), 0, "")
+	check_lenstr(xsc.Frag(1, None, 2, None, 3, None, 4), 4, "1234")
+	check_lenstr(xsc.Frag(1, (2, 3)), 3, "123")
+	check_lenstr(xsc.Frag(1, (None, None)), 1, "1")
 
 
 def test_append():
 	for cls in (xsc.Frag, html.div):
 		node = cls()
 		node.append(1)
-		check_lenunicode(node, 1, "1")
+		check_lenstr(node, 1, "1")
 		node.append(2)
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node.append()
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node.append(3, 4)
-		check_lenunicode(node, 4, "1234")
+		check_lenstr(node, 4, "1234")
 		node.append(None)
-		check_lenunicode(node, 4, "1234")
+		check_lenstr(node, 4, "1234")
 		node.append((5, 6))
-		check_lenunicode(node, 6, "123456")
+		check_lenstr(node, 6, "123456")
 		node.append(html.p.Attrs.id(7))
-		check_lenunicode(node, 7, "1234567")
+		check_lenstr(node, 7, "1234567")
 		with pytest.raises(TypeError):
 			node.append(xml.Attrs(lang=8))
 
@@ -67,61 +68,61 @@ def test_extend():
 	for cls in (xsc.Frag, html.div):
 		node = cls()
 		node.extend([1])
-		check_lenunicode(node, 1, "1")
+		check_lenstr(node, 1, "1")
 		node.extend([2])
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node.extend([])
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node.extend([None])
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node.extend([3, 4])
-		check_lenunicode(node, 4, "1234")
+		check_lenstr(node, 4, "1234")
 		node.extend([[], [[], [5], []]])
-		check_lenunicode(node, 5, "12345")
+		check_lenstr(node, 5, "12345")
 
 
 def test_insert():
 	for cls in (xsc.Frag, html.div):
 		node = cls()
 		node.insert(0, 1)
-		check_lenunicode(node, 1, "1")
+		check_lenstr(node, 1, "1")
 		node.insert(0, 2)
-		check_lenunicode(node, 2, "21")
+		check_lenstr(node, 2, "21")
 		node.insert(0, 3, 4)
-		check_lenunicode(node, 4, "3421")
+		check_lenstr(node, 4, "3421")
 		node.insert(0, None)
-		check_lenunicode(node, 4, "3421")
+		check_lenstr(node, 4, "3421")
 		node.insert(0, (5, 6))
-		check_lenunicode(node, 6, "563421")
+		check_lenstr(node, 6, "563421")
 
 
 def test_iadd():
 	for cls in (xsc.Frag, html.div):
 		node = cls()
 		node += [1]
-		check_lenunicode(node, 1, "1")
+		check_lenstr(node, 1, "1")
 		node += [2]
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node += []
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node += [None]
-		check_lenunicode(node, 2, "12")
+		check_lenstr(node, 2, "12")
 		node += [3, 4]
-		check_lenunicode(node, 4, "1234")
+		check_lenstr(node, 4, "1234")
 		node += [[], [[], [5], []]]
-		check_lenunicode(node, 5, "12345")
+		check_lenstr(node, 5, "12345")
 
 
 def test_len():
 	for cls in (xsc.Frag, html.div):
-		check_lenunicode(cls(), 0, "")
-		check_lenunicode(cls(1), 1, "1")
-		check_lenunicode(cls(1, 2, 3), 3, "123")
-		check_lenunicode(cls(None), 0, "")
-		check_lenunicode(cls(None, None, None), 0, "")
-		check_lenunicode(cls(1, None, 2, None, 3, None, 4), 4, "1234")
-		check_lenunicode(cls(1, (2, 3)), 3, "123")
-		check_lenunicode(cls(1, (None, None)), 1, "1")
+		check_lenstr(cls(), 0, "")
+		check_lenstr(cls(1), 1, "1")
+		check_lenstr(cls(1, 2, 3), 3, "123")
+		check_lenstr(cls(None), 0, "")
+		check_lenstr(cls(None, None, None), 0, "")
+		check_lenstr(cls(1, None, 2, None, 3, None, 4), 4, "1234")
+		check_lenstr(cls(1, (2, 3)), 3, "123")
+		check_lenstr(cls(1, (None, None)), 1, "1")
 
 
 def test_standardmethods():
@@ -161,11 +162,10 @@ def test_astext():
 
 def test_number():
 	node = html.div(class_=1234)
-	assert int(node["class_"]) == 1234
-	assert int(node["class_"]) == 1234
-	assert abs(float(node["class_"]) - 1234.) < 1e-2
+	assert int(node["class"]) == 1234
+	assert abs(float(node["class"]) - 1234.) < 1e-2
 	node = html.div(class_="1+1j")
-	compl = complex(node["class_"])
+	compl = complex(node["class"])
 	assert abs(compl.real - 1.) < 1e-2
 	assert abs(compl.imag - 1.) < 1e-2
 
@@ -305,15 +305,6 @@ def test_attrsclone():
 
 def test_attributes():
 	node = html.h1("gurk", {xml.Attrs.lang: "de"}, lang="de")
-	assert node.attrs.has("lang")
-	assert node.attrs.has_xml("lang")
-
-	assert node.attrs.has(html.h1.Attrs.lang)
-	assert node.attrs.has_xml(html.h1.Attrs.lang)
-
-	assert node.attrs.has(xml.Attrs.lang)
-	assert node.attrs.has_xml(xml.Attrs.lang)
-
 	assert "lang" in node.attrs
 	assert html.h1.Attrs.lang in node.attrs
 	assert xml.Attrs.lang in node.attrs
@@ -321,10 +312,10 @@ def test_attributes():
 
 def test_attributekeysvaluesitems():
 	def check(node, attrclass, attrvalue):
-		assert list(node.attrs.allowedattrs()) == [attrclass]
+		assert list(node.attrs.declaredattrs()) == [attrclass]
 
 		if attrvalue:
-			assert list(node.attrs.keys()) == [attrclass]
+			assert list(node.attrs.keys()) == [(attrclass.xmlns, attrclass.xmlname)]
 		else:
 			assert list(node.attrs.keys()) == []
 
@@ -340,7 +331,7 @@ def test_attributekeysvaluesitems():
 		if attrvalue:
 			res = list(node.attrs.items())
 			assert len(res) == 1
-			assert res[0][0] is attrclass
+			assert res[0][0] == (attrclass.xmlns, attrclass.xmlname)
 			assert res[0][1].__class__ is attrclass
 			assert str(res[0][1]) == attrvalue
 		else:
@@ -379,35 +370,21 @@ def test_attributeswithoutnames():
 		dir="ltr"
 	)
 	keys = set(node.attrs.keys())
-	keys.remove(html.h1.Attrs.class_)
+	keys.remove((None, "class"))
 
-	keys1 = set(node.attrs.withoutnames("class_").keys())
+	keys1 = set(node.attrs.withoutnames("class").keys())
 	assert keys == keys1
 
-	keys.remove(xml.Attrs.lang)
-	keys.remove(xml.Attrs.base)
-	keys2 = set(node.attrs.withoutnames("class_", xml.Attrs.lang, xml.Attrs.base).keys())
+	keys.remove((xml.xmlns, "lang"))
+	keys.remove((xml.xmlns, "base"))
+	keys2 = set(node.attrs.withoutnames("class", xml.Attrs.lang, xml.Attrs.base).keys())
 	assert keys == keys2
 
 	# Check that non existing attrs are handled correctly
-	keys3 = set(node.attrs.withoutnames("class_", "src", xml.Attrs.lang, xml.Attrs.base).keys())
+	keys3 = set(node.attrs.withoutnames("class", "src", xml.Attrs.lang, xml.Attrs.base).keys())
 	assert keys == keys3
 
 
-def test_attributeswithoutnames_xml():
-	node = html.h1(
-		"gurk",
-		title="gurk",
-		class_="important",
-		id=42,
-	)
-	keys = set(node.attrs.keys())
-	keys.remove(html.h1.Attrs.class_)
-
-	keys1 = set(node.attrs.withoutnames_xml("class").keys())
-	assert keys == keys1
-
-
 def test_attributeswithnames():
 	class h1(html.h1):
 		class Attrs(html.h1.Attrs):
@@ -423,13 +400,13 @@ def test_attributeswithnames():
 
 	assert set(node.attrs.withnames("id").keys()) == set()
 
-	assert set(node.attrs.withnames("class_").keys()) == {html.h1.Attrs.class_}
+	assert set(node.attrs.withnames("class").keys()) == {(None, "class")}
 
-	assert set(node.attrs.withnames("lang", "align").keys()) == {h1.Attrs.lang, html.h1.Attrs.align}
+	assert set(node.attrs.withnames("lang", "align").keys()) == {(None, "lang"), (None, "align")}
 
-	assert set(node.attrs.withnames(h1.Attrs.lang, "align").keys()) == {h1.Attrs.lang, html.h1.Attrs.align}
+	assert set(node.attrs.withnames(h1.Attrs.lang, "align").keys()) == {(None, "lang"), (None, "align")}
 
-	assert set(node.attrs.withnames(html.h1.Attrs.lang, "align").keys()) == {h1.Attrs.lang, html.h1.Attrs.align}
+	assert set(node.attrs.withnames(html.h1.Attrs.lang, "align").keys()) == {(None, "lang"), (None, "align")}
 
 	node = html.h1(
 		"gurk",
@@ -441,27 +418,14 @@ def test_attributeswithnames():
 
 	assert set(node.attrs.withnames("id").keys()) == set()
 
-	assert set(node.attrs.withnames("class_").keys()) == {html.h1.Attrs.class_}
+	assert set(node.attrs.withnames("class").keys()) == {(None, "class")}
 
-	assert set(node.attrs.withnames("lang", "align").keys()) == {html.h1.Attrs.lang, html.h1.Attrs.align}
+	assert set(node.attrs.withnames("lang", "align").keys()) == {(None, "lang"), (None, "align")}
 
-	# no h1.Attrs.lang
-	assert set(node.attrs.withnames(h1.Attrs.lang, "align").keys()) == {html.h1.Attrs.align}
-
-	assert set(node.attrs.withnames(html.h1.Attrs.lang, "align").keys()) == {html.h1.Attrs.lang, html.h1.Attrs.align}
-
-
-def test_attributeswithnames_xml():
-	node = html.h1(
-		"gurk",
-		{xml.Attrs.space: 1},
-		lang="de",
-		class_="gurk",
-		align="right"
-	)
-	assert set(node.attrs.withnames_xml("class").keys()) == {html.h1.Attrs.class_}
-	assert set(node.attrs.withnames_xml(xml.Attrs.space).keys()) == {xml.Attrs.space}
+	assert set(node.attrs.withnames(html.h1.Attrs.lang, "align").keys()) == {(None, "lang"), (None, "align")}
 
+	# The actual attribute class is irrelevant, ``xmlname`` and ``xmlns`` determine the attributes identity
+	assert set(node.attrs.withnames(h1.Attrs.lang, "align").keys()) == {(None, "lang"), (None, "align")}
 
 
 def test_defaultattributes():
@@ -472,8 +436,7 @@ def test_defaultattributes():
 	node = Test()
 	assert "withdef" in node.attrs
 	assert "withoutdef" not in node.attrs
-	with pytest.raises(xsc.IllegalAttrError):
-		"illegal" in node.attrs
+	assert "undeclared" not in node.attrs
 	node = Test(withdef=None)
 	assert "withdef" not in node.attrs
 
@@ -498,7 +461,7 @@ def test_attributedictmethods():
 	node = Test(withoutdef=42)
 
 	check(
-		[ Test.Attrs.withdef, Test.Attrs.withoutdef ],
+		[ (None, "withdef"), (None, "withoutdef") ],
 		list(node.attrs.keys()),
 	)
 	check(
@@ -506,13 +469,13 @@ def test_attributedictmethods():
 		list(node.attrs.values()),
 	)
 	check(
-		[ (Test.Attrs.withdef, Test.Attrs.withdef(42)), (Test.Attrs.withoutdef, Test.Attrs.withoutdef(42)) ],
+		[ ((None, "withdef"), Test.Attrs.withdef(42)), ((None, "withoutdef"), Test.Attrs.withoutdef(42)) ],
 		list(node.attrs.items()),
 	)
 
 	check(
 		[ Test.Attrs.another, Test.Attrs.withdef, Test.Attrs.withoutdef ],
-		node.attrs.allowedattrs(),
+		node.attrs.declaredattrs(),
 	)
 
 
@@ -548,7 +511,7 @@ def test_fragattrdefault():
 	assert "testattr" not in node.conv().attrs
 
 
-def test_checkisallowed():
+def test_isdeclared():
 	class testelem(xsc.Element):
 		class Attrs(xsc.Element.Attrs):
 			class testattr(xsc.TextAttr):
@@ -567,20 +530,32 @@ def test_checkisallowed():
 			testattr = None
 
 	node = testelem()
-	assert node.attrs.isallowed("testattr") is True
-	assert node.attrs.isallowed("notestattr") is False
+	assert node.attrs.isdeclared("testattr")
+	assert not node.attrs.isdeclared("notestattr")
 
 	node = testelem2()
-	assert node.attrs.isallowed("testattr") is True
-	assert node.attrs.isallowed("notestattr") is False
+	assert node.attrs.isdeclared("testattr")
+	assert not node.attrs.isdeclared("notestattr")
 
 	node = testelem3()
-	assert node.attrs.isallowed("testattr") is True
-	assert node.attrs.isallowed("testattr3") is True
+	assert node.attrs.isdeclared("testattr")
+	assert node.attrs.isdeclared("testattr3")
 
 	node = testelem4()
-	assert node.attrs.isallowed("testattr") is False
-	assert node.attrs.isallowed("testattr3") is True
+	assert not node.attrs.isdeclared("testattr")
+	assert node.attrs.isdeclared("testattr3")
+
+	# Check XML/HTML attributes
+	assert html.a.Attrs.isdeclared("href")
+	assert not html.a.Attrs.isdeclared("gurk")
+	assert not html.a.Attrs.isdeclared(xml.Attrs.lang)
+	assert xml.Attrs.isdeclared("{http://www.w3.org/XML/1998/namespace}lang")
+	assert xml.Attrs.isdeclared(xml.Attrs.lang)
+
+	# Check inherited attributes
+	assert htmlspecials.plaintable.Attrs.isdeclared("border")
+	assert htmlspecials.plaintable.Attrs.isdeclared(htmlspecials.plaintable.Attrs.border)
+	assert html.table.Attrs.isdeclared(htmlspecials.plaintable.Attrs.border)
 
 
 def test_withsep():
@@ -593,18 +568,6 @@ def test_withsep():
 		assert str(node.withsep(",")) == ""
 
 
-def test_allowedattr():
-	assert html.a.Attrs.allowedattr("href") is html.a.Attrs.href
-	with pytest.raises(xsc.IllegalAttrError):
-		html.a.Attrs.allowedattr("gurk")
-	assert html.a.Attrs.allowedattr(xml.Attrs.lang) is xml.Attrs.lang
-
-	# Check inherited attributes
-	assert htmlspecials.plaintable.Attrs.allowedattr("border") is htmlspecials.plaintable.Attrs.border
-	assert htmlspecials.plaintable.Attrs.allowedattr(htmlspecials.plaintable.Attrs.border) is htmlspecials.plaintable.Attrs.border
-	assert html.table.Attrs.allowedattr(htmlspecials.plaintable.Attrs.border) is html.table.Attrs.border
-
-
 def test_plaintableattrs():
 	e = htmlspecials.plaintable(border=3)
 	assert isinstance(e["border"], html.table.Attrs.border)
@@ -697,7 +660,7 @@ def test_getitem():
 				next(i)
 
 		# list
-		for attr in ("class_", xml.Attrs.lang):
+		for attr in ("class", xml.Attrs.lang):
 			node = cls(html.div("foo", html.div("bar", {attr: "gurk"}), "baz"))
 			assert node[[]] is node
 			assert str(node[[0, 1]]) == "bar"
@@ -897,7 +860,7 @@ def test_with():
 
 	with xsc.build():
 		with html.p() as e:
-			xsc.add(dict(class_="foo"))
+			xsc.add({"class": "foo"})
 	assert e == html.p(class_="foo")
 
 	with xsc.build():
@@ -918,3 +881,66 @@ def test_with_addattr():
 			with xsc.addattr(html.ul.Attrs.id):
 				+xsc.Text("gurk")
 	assert e == html.ul(id="gurk")
+
+
+def test_plainelement_convert():
+	node = xsc.element("x", "foo")
+
+	node = node.conv()
+
+	assert node.__class__ is xsc.Element
+	assert (node.xmlns, node.xmlname) == ("x", "foo")
+
+
+def test_plainentity_convert():
+	node = xsc.entity("foo")
+
+	node = node.conv()
+
+	assert node.__class__ is xsc.Entity
+	assert node.xmlname == "foo"
+
+
+def test_plainprocinst_convert():
+	node = xsc.procinst("foo")
+
+	node = node.conv()
+
+	assert node.__class__ is xsc.ProcInst
+	assert node.xmlname == "foo"
+
+
+def test_plainelement_mapped():
+	node = xsc.element("x", "foo")
+
+	def donothing(node, converter):
+		return node
+
+	node = node.mapped(donothing)
+
+	assert node.__class__ is xsc.Element
+	assert (node.xmlns, node.xmlname) == ("x", "foo")
+
+
+def test_plainentity_mapped():
+	node = xsc.entity("foo")
+
+	def donothing(node, converter):
+		return node
+
+	node = node.mapped(donothing)
+
+	assert node.__class__ is xsc.Entity
+	assert node.xmlname == "foo"
+
+
+def test_plainprocinst_mapped():
+	node = xsc.procinst("foo")
+
+	def donothing(node, converter):
+		return node
+
+	node = node.mapped(donothing)
+
+	assert node.__class__ is xsc.ProcInst
+	assert node.xmlname == "foo"
